/*
 * dstumbler v0.3 [aps.c]
 * by h1kari - (c) Dachb0den Labs 2001
 */

/*
 * Copyright (c) 2001 Dachb0den Labs.
 *      David Hulton <h1kari@dachb0den.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by David Hulton.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY David Hulton AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL David Hulton OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <errno.h>

#ifdef __OpenBSD__
#include <curses.h>
#else
#include <ncurses.h>
#endif

#include "dstumbler.h"
#include "screen.h"

/*
 * poll the device to detect any aps in the area
 */
void
poll_aps(const char *iface)
{
  int c, i, r, found, max_fd = 0, wep = 0;
  struct aps_s new;
  struct timeval tout, now, then;
  fd_set rset;

  /* calculate the max fd first, so we dont repeatedly do it in the loop */
  max_fd = STDIN_FILENO;

  if(usegps)
    max_fd = MAX(max_fd, gps_fd);

  for(i = 0;; i++)
  {
    apchange = 0;
    apnew = 0;

    /*  
     * now every other poll, switch to wep mode to see if there's anything
     * wep'ed out there
     */
    if(!prism2 && !nodevice)
      wep = (i % 2);

    /*
     * even if we're using a prism2 card we still need to reset the crypt bit
     * to make sure it picks up any new aps in the vicinity (thanks pete)
     */
    setcrypt(iface, &wep);

    /*
     * half the poll speed if we're dealing with a non-prism2 and have to poll
     * for both weped and non-weped
     */
    gettimeofday(&now, NULL);
    then.tv_usec = now.tv_usec + (prism2 ? POLLSPEED : POLLSPEED / 2);
    then.tv_sec = now.tv_sec + (then.tv_usec / 1000000);
    then.tv_usec %= 1000000;

    /*
     * now loop with the now and then to guarantee we wait for the specified
     * timeout.
     */
    while(now.tv_sec < then.tv_sec ? 1 :
     (now.tv_sec == then.tv_sec ? now.tv_usec < then.tv_usec : 0))
    {
      FD_ZERO(&rset);
      FD_SET(STDIN_FILENO, &rset);

      if(usegps)
        FD_SET(gps_fd, &rset);

      tout.tv_sec = then.tv_sec - now.tv_sec;
      tout.tv_usec = then.tv_usec - now.tv_usec;

      /* if our usec is < 0, lets borrow from the seconds so it's positive */
      if(tout.tv_usec < 0)
      {
        tout.tv_sec--;
        tout.tv_usec = 1000000 + tout.tv_usec;
      }

      if((r = select(max_fd + 1, &rset, NULL, NULL, &tout)) == -1)
      {
        alert("error: error with select(): %s", strerror(errno));
        exitclean(2);
      }

      if(r == 0)
        break;
      else if(FD_ISSET(STDIN_FILENO, &rset) && (c = getch()))
        parse_cmd(c);
      else if(usegps && FD_ISSET(gps_fd, &rset))
        gps_parse(gps_fd);

      gettimeofday(&now, NULL);
    }

    if(nodevice)
    {
      smart_redraw();
      continue;
    }

    /* first make sure we're connected to something */
    getbssid(iface, new.bssid);
    getquality(iface, new.quality);

    if((new.bssid[0] == 0x44 && new.bssid[1] == 0x44 && new.bssid[2] == 0x44)
     || (!prism2 && new.quality[1] == 27 && new.quality[2] == 27) ||
     (prism2 && new.quality[1] == 47))
    {
      smart_redraw();
      continue;
    }

    getssid(iface, new.ssid);
    getchan(iface, &new.chan);

    /* get cached gps coordinates from select loop */
    if(usegps)
    {
      new.ns.dir = ns.dir;
      new.ns.coordinates = ns.coordinates;
      new.ew.dir = ew.dir;
      new.ew.coordinates = ew.coordinates;
    }
    else
    {
      new.ns.dir = 'N';
      new.ns.coordinates = 0;
      new.ew.dir = 'E';
      new.ew.coordinates = 0;
    }

    /* manually calculate quality (s/n ratio) if we're using a prism2 */
    if(prism2)
    {
      new.quality[2] = new.quality[0];

      /* default noise is 27, if we get it assume it isn't accurate */
      new.quality[0] = (new.quality[0] == 27 && 
        new.quality[1] < new.quality[2] ? 0 :
        new.quality[1] - new.quality[2]);
    }

    if((found = find_ap(new.ssid, new.bssid, new.chan)) == -1)
    {
      /* set timestamps */
      gettimeofday(&new.firstseen, NULL);
      gettimeofday(&new.lastseen, NULL);

      /* set wep mode */
      new.weped = wep;

      add_ap(&new);

      redraw_aps();

      /* if we're in autoselect mode, make sure we refresh stats for this ap */
      if(autosel)
      {
        redraw_apinfo();
        redraw_graph();
      }
    }
    /* just redraw (if necessary) and continue if we don't get anything */
    else if(found == -2)
    {
      smart_redraw();
      continue;
    }
    else
    {
      if(!wep && !prism2)
        aps[found]->weped = wep;

      gettimeofday(&new.lastseen, NULL);

      /* update the found entry with quality info and timestamp and gps */
      aps[found]->quality[0] = new.quality[0];
      aps[found]->quality[1] = new.quality[1];
      aps[found]->quality[2] = new.quality[2];

      aps[found]->lastseen.tv_sec = new.lastseen.tv_sec;
      aps[found]->lastseen.tv_usec = new.lastseen.tv_usec;

      aps[found]->ns.dir = new.ns.dir;
      aps[found]->ns.coordinates = new.ns.coordinates;

      aps[found]->ew.dir = new.ew.dir;
      aps[found]->ew.coordinates = new.ew.coordinates;

      /* write log */
      add_log(aps[found], &new);

      if(found != aps_new || autosel)
        APNEW(found);

      redraw_aps();

      if(found == aps_cur || autosel)
      {
        redraw_apinfo();
        redraw_graph();
      }
    }

    memset((char *)&new, 0, sizeof(struct aps_s));
    smart_redraw();
  }
}

/*
 * search through the aps linked list to see if the ap already exists
 */
int
find_ap(const char *ssid, const char *bssid, int chan)
{
  int i, found = -1;

  for(i = 0; i < aps_len; i++)
  {
    if(memcmp(aps[i]->bssid, bssid, MACSIZE) == 0 &&
     (strcmp(aps[i]->ssid, ssid) == 0 ||
     /* if we can't match ssid, make sure can at least match channel */
     (strlen(ssid) == 0 && aps[i]->chan == chan)))
      found = i;

     /* if this isn't the ap we're looking for, zero out it's quality info */
     memset((char *)aps[i]->quality, 0, sizeof(aps[i]->quality));
  }

  if(found == -1 && strlen(ssid) == 0)
    return -2;

  return found;
}

/*
 * redraw the ap screen with all of the aps in the aps linked list
 */
#define TEMP_SSID_LEN (AP_SCR_COLS - 42)
#define SSID_FMT_LEN 7
#define MFG_FMT_LEN 7
#define MAC_STR_LEN 17
void
redraw_aps(void)
{
  int off = 0, i = 0, wrote = 0, ap_color, ap_color_bold, code;
  char temp_ssid[TEMP_SSID_LEN], ssid_fmt[SSID_FMT_LEN];
  struct aps_s *ptr;

  werase(ap_scr);
  off = apchange ? aps_cur : aps_new;

  if(off < aps_win)
    aps_win = off;
  else if(off > (aps_win + AP_SCR_LINES - 1))
    aps_win = off - (AP_SCR_LINES - 1);

  for(i = aps_win; i < aps_len; i++)
  {
    ptr = aps[i];

    if(wrote++)
      waddch(ap_scr, '\n');

    if(i == aps_cur)
    {
      ap_color = AP_COLOR_SEL;
      ap_color_bold = AP_COLOR_SEL_BOLD;
    }
    else
    {
      ap_color = AP_COLOR;
      ap_color_bold = AP_COLOR_BOLD;
    }

    memset(temp_ssid, 0, TEMP_SSID_LEN - 1);
    memcpy(temp_ssid, ptr->ssid, MIN(SSIDLEN, (TEMP_SSID_LEN - 1)));
    temp_ssid[TEMP_SSID_LEN - 1] = '\0';
    snprintf(ssid_fmt, SSID_FMT_LEN - 1, "%%-%ds", TEMP_SSID_LEN - 1);

    /* figure out how many symbols our quality is to determine color code */
    if(ptr->quality[0] < 4)
      code = COLOR_CODE_LOW;
    else if(ptr->quality[0] > 5 && ptr->quality[0] < 9)
      code = COLOR_CODE_LOW_BOLD;
    else if(ptr->quality[0] > 8 && ptr->quality[0] < 12)
      code = COLOR_CODE_MED;
    else if(ptr->quality[0] > 11 && ptr->quality[0] < 15)
      code = COLOR_CODE_MED_BOLD;
    else if(ptr->quality[0] > 14 && ptr->quality[0] < 18)
      code = COLOR_CODE_HI;
    else if(ptr->quality[0] > 17)
      code = COLOR_CODE_HI_BOLD;

    wattrset(ap_scr, 0);

    /*
     * sorry for the messyness... better then using a loop
     */
    waddch(ap_scr, ((autosel ? i == aps_new : i == aps_cur) ? APS_CUR | code :
     (i == aps_new ? APS_NEW | code : ' ')));

    wattrset(ap_scr, ap_color_bold);
    wprintw(ap_scr, " [");
    wattrset(ap_scr, ap_color);
    wprintw(ap_scr, "%2d", ptr->chan);
    wattrset(ap_scr, ap_color_bold);
    wprintw(ap_scr, "] ");
    wattrset(ap_scr, ap_color);
    wprintw(ap_scr, ssid_fmt, temp_ssid);
    wattrset(ap_scr, ap_color_bold);
    wprintw(ap_scr, " (");
    wattrset(ap_scr, ap_color);

    if(resolvmfg)
    {
      char mfg[MFG_STRLEN], mfg_fmt[MFG_FMT_LEN];
      mfgresolve(mfg, ptr->bssid);
      snprintf(mfg_fmt, MFG_FMT_LEN - 1, "%%-%ds", MAC_STR_LEN);
      wprintw(ap_scr, mfg_fmt, mfg);
    }
    else
      wprintw(ap_scr, "%02x:%02x:%02x:%02x:%02x:%02x",
       ptr->bssid[0], ptr->bssid[1], ptr->bssid[2],
       ptr->bssid[3], ptr->bssid[4], ptr->bssid[5]);

    wattrset(ap_scr, ap_color_bold);
    wprintw(ap_scr, ") ");
    waddch(ap_scr, ptr->weped ? APS_WEP | AP_COLOR_WEPED : ' ');
    wattrset(ap_scr, ap_color);
    wprintw(ap_scr, "%03d", ptr->quality[0]);
    waddch(ap_scr, ':' | ap_color_bold);
    wprintw(ap_scr, "%03d", ptr->quality[1]);
    waddch(ap_scr, ':' | ap_color_bold);
    wprintw(ap_scr, "%03d", ptr->quality[2]);

    if(wrote == AP_SCR_LINES)
      break;
  }

  redraw.ap_scr++;
}

/*
 * redraw current ap's info (in case we've polled more info from it)
 */
#define INFO_BUF_LEN (INFO_SCR_COLS * INFO_SCR_LINES)
#define MFG_BUF_LEN INFO_SCR_COLS
#define FIRSTSEEN_LEN 10
#define LASTSEEN_LEN 10
void
redraw_apinfo(void)
{
  int i, mode = 0;
  char info_buf[INFO_BUF_LEN], firstseen[FIRSTSEEN_LEN], lastseen[LASTSEEN_LEN];
  char mfg_buf[MFG_BUF_LEN], mfg[MFG_STRLEN];
  struct aps_s *ptr;

  ptr = aps[autosel && apnew ? aps_new : aps_cur];

  strftime(firstseen, FIRSTSEEN_LEN - 1, "%H:%M:%S",
   localtime((time_t *)&ptr->firstseen.tv_sec));
  strftime(lastseen, LASTSEEN_LEN - 1, "%H:%M:%S",
   localtime((time_t *)&ptr->lastseen.tv_sec));

  if(usegps)
    snprintf(mfg_buf, MFG_BUF_LEN - 1, "GPS: %c:%.3f %c:%.3f",
     ptr->ns.dir, ptr->ns.coordinates, ptr->ew.dir, ptr->ew.coordinates);
  else
  {
    mfgresolve(mfg, ptr->bssid);
    snprintf(mfg_buf, MFG_BUF_LEN - 1, "Mfg: %s", mfg);
  }

  werase(info_scr);
  snprintf(info_buf, INFO_BUF_LEN - 1,
   "SSID: %s\n"
   "BSSID: %02x:%02x:%02x:%02x:%02x:%02x\n"
   "%s\n"
   "Channel: %d\n"
   "Signal/Noise: %d/%d/%d\n"
   "First Seen: %s\n"
   "Last Seen:  %s\n",
   ptr->ssid, ptr->bssid[0], ptr->bssid[1], ptr->bssid[2], ptr->bssid[3],
   ptr->bssid[4], ptr->bssid[5], mfg_buf, ptr->chan, ptr->quality[0],
   ptr->quality[1], ptr->quality[2], firstseen, lastseen);

  for(i = 0; i < strlen(info_buf); i++)
  {
    if(info_buf[i] == ':' && info_buf[i + 1] == ' ')
      mode = 1;
    else if(info_buf[i] == '\n')
      mode = 0;

    if(mode && info_buf[i] != ':')
      waddch(info_scr, info_buf[i] | INFO_COLOR_BOLD);
    else
      waddch(info_scr, info_buf[i] | INFO_COLOR);
  }

  redraw.info_scr++;
}

/*
 * redraw signal to noise graph for current ap
 */
void
redraw_graph(void)
{
  static int wrote = 0;
  int i;
  struct aps_s *ptr;

  ptr = aps[autosel ? aps_new : aps_cur];

  if(apchange || (autosel && apnew))
  {
    werase(graph_scr);
    wrote = 0;

    for(i = (ptr->log_len > GRAPH_SCR_LINES ?
     ptr->log_len - GRAPH_SCR_LINES : 0); i < ptr->log_len; i++)
      print_graph(ptr->log[i]->quality, wrote++);
  }
  else
    print_graph(ptr->quality, wrote++);

  redraw.graph_scr++;
}

/*
 * print one line on the graph using quality array
 */
void
print_graph(int *quality, int wrote)
{
  int i, s2n[2], sym[2], max = (prism2 ? GRAPH_PRISM_MAX : GRAPH_NORM_MAX);

  for(i = 0; i < 2; i++)
    s2n[i] = ((GRAPH_SCR_COLS - 12) * quality[i + 1]) / max;

  sym[0] = GRAPH_SIGNAL | GRAPH_COLOR_SIGNAL;
  sym[1] = GRAPH_NOISE | GRAPH_COLOR_NOISE;

  /* in case noise is > signal */
  if(s2n[1] > s2n[0])
  {
    SWAP(s2n[0], s2n[1]);
    SWAP(sym[0], sym[1]);
  }

  if(wrote)
    waddch(graph_scr, '\n');

  wprintw(graph_scr, "%03d", quality[0]);
  waddch(graph_scr, ':' | GRAPH_COLOR_BOLD);
  wprintw(graph_scr, "%03d", quality[1]);
  waddch(graph_scr, ':' | GRAPH_COLOR_BOLD);
  wprintw(graph_scr, "%03d ", quality[2]);

  for(i = 0; i < (GRAPH_SCR_COLS - 13); i++)
  {
    if(i < s2n[1])
      waddch(graph_scr, sym[1]);
    else if(i < s2n[0])
      waddch(graph_scr, sym[0]);
    else
      waddch(graph_scr, ' ');
  }
}

/*
 * add ap to the linked list (along with a log of the addition)
 */
#define LOG_INIT_LEN (sizeof(struct log_s *) * PREALLOC_LOG)
void
add_ap(const struct aps_s *ap)
{
  int i;
  struct aps_s *ptr;

  /* reallocate aps pointer array if we're reaching the end */
  if(aps_len == aps_max)
  {
    struct aps_s **new_aps;

    if((new_aps = (struct aps_s **)realloc(aps, (aps_max + PREALLOC_APS) *
     sizeof(struct aps_s *))) == NULL)
    {
      alert("error: unable to reallocate memory: %s", strerror(errno));
      exitclean(2);
    }

    aps = new_aps;

    for(i = aps_max; i < (aps_max + PREALLOC_APS); i++)
      aps[i] = NULL;

    aps_max += PREALLOC_APS;
  }

  if((aps[aps_len] = (struct aps_s *)malloc(sizeof(struct aps_s))) == NULL)
  {
    alert("error: unable to allocate memory: %s", strerror(errno));
    exitclean(2);
  }

  memcpy((char *)aps[aps_len], (char *)ap, sizeof(struct aps_s));

  ptr = aps[aps_len];

  /* now pre-allocate log pointer array */
  if((ptr->log = (struct log_s **)malloc(LOG_INIT_LEN))
   == NULL)
  {
    alert("error: unable to allocate memory: %s", strerror(errno));
    exitclean(2);
  }

  memset((char *)ptr->log, 0, LOG_INIT_LEN);
  ptr->log_max = PREALLOC_LOG;
  ptr->log_len = 0;

  add_log(ptr, ap);
  APNEW(aps_len);

  aps_len++;
}

/*
 * add log to the specified ap
 */
void
add_log(struct aps_s *ap, const struct aps_s *info)
{
  int i;
  struct log_s *logptr;

  /* reallocate log pointer array if we're reaching the end */
  if(ap->log_len == ap->log_max)
  {
    struct log_s **new_log;

    if((new_log = (struct log_s **)realloc(ap->log,
     (ap->log_max + PREALLOC_LOG) * sizeof(struct log_s *))) == NULL)
    {
      alert("error: unable to reallocate memory: %s", strerror(errno));
      exitclean(2);
    }

    ap->log = new_log;

    for(i = ap->log_max; i < (ap->log_max + PREALLOC_LOG); i++)
      ap->log[i] = NULL;

    ap->log_max += PREALLOC_LOG;
  }

  /* make first log entry */
  if((ap->log[ap->log_len] = (struct log_s *)malloc(sizeof(struct log_s)))
   == NULL)
  {
    alert("error: unable to allocate memory: %s", strerror(errno));
    exitclean(2);
  }

  memset((char *)ap->log[ap->log_len], 0, sizeof(struct log_s));

  logptr = ap->log[ap->log_len];

  logptr->ns.dir = info->ns.dir;
  logptr->ns.coordinates = info->ns.coordinates;

  logptr->ew.dir = info->ew.dir;
  logptr->ew.coordinates = info->ew.coordinates;

  logptr->quality[0] = info->quality[0];
  logptr->quality[1] = info->quality[1];
  logptr->quality[2] = info->quality[2];

  logptr->seen.tv_sec = info->lastseen.tv_sec;
  logptr->seen.tv_usec = info->lastseen.tv_usec; 

  ap->log_len++;
}
