/*
 * dstumbler v0.3 [wistat.c]
 * by h1kari - (c) Dachb0den Labs 2001
 */

/*
 * Copyright (c) 2001 Dachb0den Labs.
 *      David Hulton <h1kari@dachb0den.com>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by David Hulton.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY David Hulton AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL David Hulton OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>

#ifdef __OpenBSD__
#include <curses.h>
#else
#include <ncurses.h>
#endif

#include "dstumbler.h"

#define wi_getval(x, y) wi_ioctl(x, y, SIOCGWAVELAN)
#define wi_setval(x, y) wi_ioctl(x, y, SIOCSWAVELAN)

/*
 * calls ioctl on a socket to the device to poll information
 */
void
wi_ioctl(const char *iface, struct wi_req *wreq, int cmd)
{
  static int s = -1;
  struct ifreq ifr;

  memset((char *)&ifr, 0, sizeof(ifr));
  strcpy(ifr.ifr_name, iface);
  ifr.ifr_data = (caddr_t)wreq;

  if(s < 0 ? (s = socket(AF_INET, SOCK_DGRAM, 0)) == -1 : 0)
  {
    alert("error: unable to create socket: %s", strerror(errno));
    exitclean(2);
  }

  if(ioctl(s, cmd, &ifr) == -1)
  {
    alert("error: unable to ioctl device socket: %s", strerror(errno));
    exitclean(2);
  }
}

/*
 * stat the card for specific info. the void * will be filled in depending
 * on the cmd. you must make sure the memory it's writing to is allocated
 * enough for the operation that's being performed.
 */
void
wi_getstat(const char *iface, int cmd, char *mem, int len)
{
  int i;
  struct wi_req wreq;

  memset((char *)&wreq, 0, sizeof(struct wi_req));
  wreq.wi_len = WI_MAX_DATALEN;
  wreq.wi_type = cmd;

  wi_getval(iface, &wreq);

  switch(cmd)
  {
    case WI_RID_CURRENT_SSID:
    case WI_RID_OWN_SSID:
      wi_strncpyval(mem, (char *)wreq.wi_val, len);
      break;
    case WI_RID_CURRENT_BSSID:
    case WI_RID_MAC_NODE:
      memcpy(mem, (char *)wreq.wi_val, len);
      break;
    case WI_RID_CURRENT_CHAN:
    case WI_RID_ENCRYPTION:
      *((int *)mem) = le16toh(wreq.wi_val[0]);
      break;
    case WI_RID_COMMS_QUALITY:
      for(i = 0; i < (wreq.wi_len - 1); i++)
        ((int *)mem)[i] = le16toh(wreq.wi_val[i]);
      break;
    case WI_RID_DEFLT_CRYPT_KEYS:
      memcpy(mem, (char *)&wreq, sizeof(struct wi_ltv_keys));
      break;
    default:
      break;
  }
}

/*
 * set stats on the card. the void * will be parsed depending on the cmd.
 */
void
wi_setstat(const char *iface, int cmd, char *mem, int len)
{
  struct wi_req wreq;

  memset((char *)&wreq, 0, sizeof(wreq));
  wreq.wi_type = cmd;

  switch(cmd)
  {
    case WI_RID_ENCRYPTION:
      wreq.wi_len = len;
      wreq.wi_val[0] = htole16(*((int *)mem));
      break;
    case WI_RID_OWN_SSID:
      wreq.wi_len = MIN((strlen(mem) + 1) / 2, len);
      wreq.wi_val[0] = htole16(strlen(mem));
      memcpy((char *)&wreq.wi_val[1], mem, strlen(mem));
      break;
    case WI_RID_DEFLT_CRYPT_KEYS:
      wreq.wi_len = len;
      memcpy((char *)&wreq.wi_val,
       (char *)&((struct wi_ltv_keys *)mem)->wi_keys,
       sizeof(struct wi_key) * 4);
      break;
    case WI_RID_MAC_NODE:
      wreq.wi_len = len / 2;
      memcpy((char *)&wreq.wi_val, mem, len);
      break;
    default:
      break;
  }

  wi_setval(iface, &wreq);
}

/*
 * decodes a string in a wi_val
 */
void
wi_strncpyval(char *dst, const char *src, int maxlen)
{
  int i, len;

  len = le16toh(*(u_int16_t *)src);
  len = MIN(len, maxlen - 1);

  memcpy(dst, (src + sizeof(u_int16_t)), len);

  /* filter out any stray \0's */
  for(i = 0; i < len; i++)
  {
    if(dst[i] == '\0')
      dst[i] = ' ';
  }

  dst[len] = '\0';
}

/*
 * start encryption stuff (set key to bogous value so we can detect weped
 * networks)
 */
void
wi_start(const char *iface)
{
  int i, zero = 0;
  u_char temp_macaddr[MACSIZE];
  struct wi_ltv_keys keys;

  /* first backup the current keys and encryption status */
  getcryptkeys(iface, &backup.keys);
  getcrypt(iface, &backup.crypt);

  memset((char *)&keys, 0, sizeof(struct wi_ltv_keys));

  /* now set our bogous info */
  for(i = 0; i < 4; i++)
  {
    keys.wi_keys[i].wi_keylen = htole16(5);
    keys.wi_keys[i].wi_keydat[0] = 0x01;
    keys.wi_keys[i].wi_keydat[1] = 0x23;
    keys.wi_keys[i].wi_keydat[2] = 0x45;
    keys.wi_keys[i].wi_keydat[3] = 0x67;
    keys.wi_keys[i].wi_keydat[4] = 0x89;
  }

  setcryptkeys(iface, &keys);
  setcrypt(iface, &zero);

  getdefaultssid(iface, backup.defaultssid);

  /* set a null ssid for the default */
  setdefaultssid(iface, "\0");

  if(macreset)
  {
    /* backup the mac address if we're told to spoof */
    getmacaddr(iface, backup.macaddr);

    /* set a random mac address */
    srand((u_int)(time(NULL) - (1 << 31)));

    temp_macaddr[0] = 0;
    for(i = 1; i < MACSIZE; i++)
      temp_macaddr[i] = rand() % 256;

    setmacaddr(iface, temp_macaddr);
  }

  backup.wi_started++;
}

/*
 * stop encryption stuff (restore previous values for the device)
 */
void
wi_stop(const char *iface)
{
  /* set the keys back to what they were */
  setcryptkeys(iface, &backup.keys);

  /* set the encryption status back */
  setcrypt(iface, &backup.crypt);

  /* set the default ssid back */
  setdefaultssid(iface, backup.defaultssid);

  if(macreset)
    /* reset the mac address back to what it was */
    setmacaddr(iface, backup.macaddr);

  backup.wi_started = 0;
}
